import hydrostats.data as hd
import pandas as pd
import plotly.graph_objects as go
import scipy.stats
from plotly.offline import plot as offline_plot

from .format_tools import build_title
from .plotly_helpers import _rperiod_scatters

__all__ = [
    'corrected_retrospective',
    'corrected_scatterplots',
    'corrected_day_average',
    'corrected_month_average',
    'corrected_volume_compare',
]


# BIAS CORRECTION PLOTS
def corrected_retrospective(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                            rperiods: pd.DataFrame = None, plot_titles: list = None,
                            plot_type: str = 'plotly') -> go.Figure or str:
    """
    Creates a plot of corrected discharge, observed discharge, and simulated discharge

    Args:
        corrected: the response from the geoglows.bias.correct_historical_simulation function\
        simulated: the csv response from historic_simulation
        observed: the dataframe of observed data. Must have a datetime index and a single column of flow values
        rperiods: the csv response from return_periods
        plot_type: either 'plotly', or 'plotly_html' (default plotly)
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}

    Returns:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    startdate = corrected.index[0]
    enddate = corrected.index[-1]

    plot_data = {
        'x_simulated': corrected.index.tolist(),
        'x_observed': observed.index.tolist(),
        'y_corrected': corrected.values.flatten(),
        'y_simulated': simulated.values.flatten(),
        'y_observed': observed.values.flatten(),
        'y_max': max(corrected.values.max(), observed.values.max(), simulated.values.max()),
    }
    if rperiods is not None:
        plot_data.update(rperiods.to_dict(orient='index').items())
        rperiod_scatters = _rperiod_scatters(startdate, enddate, rperiods, plot_data['y_max'], plot_data['y_max'])
    else:
        rperiod_scatters = []

    if plot_type == 'json':
        return plot_data

    scatters = [
        go.Scatter(
            name='Simulated Data',
            x=plot_data['x_simulated'],
            y=plot_data['y_simulated'],
            line=dict(color='red')
        ),
        go.Scatter(
            name='Observed Data',
            x=plot_data['x_observed'],
            y=plot_data['y_observed'],
            line=dict(color='blue')
        ),
        go.Scatter(
            name='Corrected Simulated Data',
            x=plot_data['x_simulated'],
            y=plot_data['y_corrected'],
            line=dict(color='#00cc96')
        ),
    ]
    scatters += rperiod_scatters

    layout = go.Layout(
        title=build_title("Historical Simulation Comparison", plot_titles),
        yaxis={'title': 'Discharge (m<sup>3</sup>/s)'},
        xaxis={'title': 'Date (UTC +0:00)', 'range': [startdate, enddate], 'hoverformat': '%b %d %Y',
               'tickformat': '%Y'},
    )

    figure = go.Figure(data=scatters, layout=layout)
    if plot_type == 'plotly':
        return figure
    if plot_type == 'plotly_html':
        return offline_plot(
            figure,
            config={'autosizable': True, 'responsive': True},
            output_type='div',
            include_plotlyjs=False
        )
    raise ValueError('Invalid plot_type chosen. Choose json, plotly, plotly_scatters, or plotly_html')


def corrected_scatterplots(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                           merged_sim_obs: pd.DataFrame = False, merged_cor_obs: pd.DataFrame = False,
                           plot_titles: list = None, ) -> go.Figure or str:
    """
    Creates a plot of corrected discharge, observed discharge, and simulated discharge. This function uses
    hydrostats.data.merge_data on the 3 inputs. If you have already computed these because you are doing a full
    comparison of bias correction, you can provide them to save time

    Args:
        corrected: the response from the geoglows.bias.correct_historical_simulation function
        simulated: the csv response from historic_simulation
        observed: the dataframe of observed data. Must have a datetime index and a single column of flow values
        merged_sim_obs: (optional) if you have already computed it, hydrostats.data.merge_data(simulated, observed)
        merged_cor_obs: (optional) if you have already computed it, hydrostats.data.merge_data(corrected, observed)
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}

    Returns:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    if corrected is False and simulated is False and observed is False:
        if merged_sim_obs is not False and merged_cor_obs is not False:
            pass  # if you provided the merged dataframes already, we use those
    else:
        # merge the datasets together
        merged_sim_obs = hd.merge_data(sim_df=simulated, obs_df=observed)
        merged_cor_obs = hd.merge_data(sim_df=corrected, obs_df=observed)

    # get the min/max values for plotting the 45 degree line
    min_value = min(min(merged_sim_obs.iloc[:, 1].values), min(merged_sim_obs.iloc[:, 0].values))
    max_value = max(max(merged_sim_obs.iloc[:, 1].values), max(merged_sim_obs.iloc[:, 0].values))

    # do a linear regression on both of the merged dataframes
    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(merged_sim_obs.iloc[:, 0].values,
                                                                         merged_sim_obs.iloc[:, 1].values)
    slope2, intercept2, r_value2, p_value2, std_err2 = scipy.stats.linregress(merged_cor_obs.iloc[:, 0].values,
                                                                              merged_cor_obs.iloc[:, 1].values)
    scatter_sets = [
        go.Scatter(
            x=merged_sim_obs.iloc[:, 0].values,
            y=merged_sim_obs.iloc[:, 1].values,
            mode='markers',
            name='Original Data',
            marker=dict(color='#ef553b')
        ),
        go.Scatter(
            x=merged_cor_obs.iloc[:, 0].values,
            y=merged_cor_obs.iloc[:, 1].values,
            mode='markers',
            name='Corrected',
            marker=dict(color='#00cc96')
        ),
        go.Scatter(
            x=[min_value, max_value],
            y=[min_value, max_value],
            mode='lines',
            name='45 degree line',
            line=dict(color='black')
        ),
        go.Scatter(
            x=[min_value, max_value],
            y=[slope * min_value + intercept, slope * max_value + intercept],
            mode='lines',
            name=f'Y = {round(slope, 2)}x + {round(intercept, 2)} (Original)',
            line=dict(color='red')
        ),
        go.Scatter(
            x=[min_value, max_value],
            y=[slope2 * min_value + intercept2, slope2 * max_value + intercept2],
            mode='lines',
            name=f'Y = {round(slope2, 2)}x + {round(intercept2, 2)} (Corrected)',
            line=dict(color='green')
        )
    ]

    updatemenus = [
        dict(active=0,
             buttons=[dict(label='Linear Scale',
                           method='update',
                           args=[{'visible': [True, True]},
                                 {'title': 'Linear scale',
                                  'yaxis': {'type': 'linear'}}]),
                      dict(label='Log Scale',
                           method='update',
                           args=[{'visible': [True, True]},
                                 {'title': 'Log scale',
                                  'xaxis': {'type': 'log'},
                                  'yaxis': {'type': 'log'}}]),
                      ]
             )
    ]

    layout = go.Layout(title=build_title('Bias Correction Scatter Plot', plot_titles),
                       xaxis=dict(title='Simulated', ),
                       yaxis=dict(title='Observed', autorange=True),
                       showlegend=True, updatemenus=updatemenus)
    return go.Figure(data=scatter_sets, layout=layout)


def corrected_month_average(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                            merged_sim_obs: pd.DataFrame = False, merged_cor_obs: pd.DataFrame = False,
                            plot_titles: list = None, ) -> go.Figure or str:
    """
    Calculates and _plots the monthly average streamflow. This function uses
    hydrostats.data.merge_data on the 3 inputs. If you have already computed these because you are doing a full
    comparison of bias correction, you can provide them to save time

    Args:
        corrected: the response from the geoglows.bias.correct_historical_simulation function
        simulated: the csv response from historic_simulation
        observed: the dataframe of observed data. Must have a datetime index and a single column of flow values
        merged_sim_obs: (optional) if you have already computed it, hydrostats.data.merge_data(simulated, observed)
        merged_cor_obs: (optional) if you have already computed it, hydrostats.data.merge_data(corrected, observed)
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}

    Returns:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    if corrected is False and simulated is False and observed is False:
        if merged_sim_obs is not False and merged_cor_obs is not False:
            pass  # if you provided the merged dataframes already, we use those
    else:
        # merge the datasets together
        merged_sim_obs = hd.merge_data(sim_df=simulated, obs_df=observed)
        merged_cor_obs = hd.merge_data(sim_df=corrected, obs_df=observed)
    monthly_avg = hd.monthly_average(merged_sim_obs)
    monthly_avg2 = hd.monthly_average(merged_cor_obs)

    scatters = [
        go.Scatter(x=monthly_avg.index, y=monthly_avg.iloc[:, 1].values, name='Observed Data'),
        go.Scatter(x=monthly_avg.index, y=monthly_avg.iloc[:, 0].values, name='Simulated Data'),
        go.Scatter(x=monthly_avg2.index, y=monthly_avg2.iloc[:, 0].values, name='Corrected Simulated Data'),
    ]

    layout = go.Layout(
        title=build_title('Monthly Average Streamflow Comparison', plot_titles=plot_titles),
        xaxis=dict(title='Month'), yaxis=dict(title='Discharge (m<sup>3</sup>/s)', autorange=True),
        showlegend=True)

    return go.Figure(data=scatters, layout=layout)


def corrected_day_average(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                          merged_sim_obs: pd.DataFrame = False, merged_cor_obs: pd.DataFrame = False,
                          plot_titles: dict = None, ) -> go.Figure or str:
    """
    Calculates and _plots the daily average streamflow. This function uses
    hydrostats.data.merge_data on the 3 inputs. If you have already computed these because you are doing a full
    comparison of bias correction, you can provide them to save time

    Args:
        corrected: the response from the geoglows.bias.correct_historical_simulation function
        simulated: the csv response from historic_simulation
        merged_sim_obs: (optional) if you have already computed it, hydrostats.data.merge_data(simulated, observed)
        merged_cor_obs: (optional) if you have already computed it, hydrostats.data.merge_data(corrected, observed)
        observed: the dataframe of observed data. Must have a datetime index and a single column of flow values
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}

    Returns:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    if corrected is False and simulated is False and observed is False:
        if merged_sim_obs is not False and merged_cor_obs is not False:
            pass  # if you provided the merged dataframes already, we use those
    else:
        # merge the datasets together
        merged_sim_obs = hd.merge_data(sim_df=simulated, obs_df=observed)
        merged_cor_obs = hd.merge_data(sim_df=corrected, obs_df=observed)
    daily_avg = hd.daily_average(merged_sim_obs)
    daily_avg2 = hd.daily_average(merged_cor_obs)

    scatters = [
        go.Scatter(x=daily_avg.index, y=daily_avg.iloc[:, 1].values, name='Observed Data'),
        go.Scatter(x=daily_avg.index, y=daily_avg.iloc[:, 0].values, name='Simulated Data'),
        go.Scatter(x=daily_avg2.index, y=daily_avg2.iloc[:, 0].values, name='Corrected Simulated Data'),
    ]

    layout = go.Layout(
        title=build_title('Daily Average Streamflow Comparison', plot_titles),
        xaxis=dict(title='Days'), yaxis=dict(title='Discharge (m<sup>3</sup>/s)', autorange=True),
        showlegend=True)

    return go.Figure(data=scatters, layout=layout)


def corrected_volume_compare(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                             merged_sim_obs: pd.DataFrame = False, merged_cor_obs: pd.DataFrame = False,
                             plot_titles: dict = None, ) -> go.Figure or str:
    """
    Calculates and _plots the cumulative volume output on each of the 3 datasets provided. This function uses
    hydrostats.data.merge_data on the 3 inputs. If you have already computed these because you are doing a full
    comparison of bias correction, you can provide them to save time

    Args:
        corrected: the response from the geoglows.bias.correct_historical_simulation function
        simulated: the csv response from historic_simulation
        observed: the dataframe of observed data. Must have a datetime index and a single column of flow values
        merged_sim_obs: (optional) if you have already computed it, hydrostats.data.merge_data(simulated, observed)
        merged_cor_obs: (optional) if you have already computed it, hydrostats.data.merge_data(corrected, observed)
        plot_titles: (dict) Extra info to show on the title of the plot. For example:
            {'River ID': 1234567, 'Drainage Area': '1000km^2'}

    Returns:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    if corrected is False and simulated is False and observed is False:
        if merged_sim_obs is not False and merged_cor_obs is not False:
            pass  # if you provided the merged dataframes already, we use those
    else:
        # merge the datasets together
        merged_sim_obs = hd.merge_data(sim_df=simulated, obs_df=observed)
        merged_cor_obs = hd.merge_data(sim_df=corrected, obs_df=observed)

    sim_array = merged_sim_obs.iloc[:, 0].values
    obs_array = merged_sim_obs.iloc[:, 1].values
    corr_array = merged_cor_obs.iloc[:, 0].values

    sim_volume_dt = sim_array * 0.0864
    obs_volume_dt = obs_array * 0.0864
    corr_volume_dt = corr_array * 0.0864

    sim_volume_cum = []
    obs_volume_cum = []
    corr_volume_cum = []
    sum_sim = 0
    sum_obs = 0
    sum_corr = 0

    for i in sim_volume_dt:
        sum_sim = sum_sim + i
        sim_volume_cum.append(sum_sim)

    for j in obs_volume_dt:
        sum_obs = sum_obs + j
        obs_volume_cum.append(sum_obs)

    for k in corr_volume_dt:
        sum_corr = sum_corr + k
        corr_volume_cum.append(sum_corr)

    observed_volume = go.Scatter(x=merged_sim_obs.index, y=obs_volume_cum, name='Observed', )
    simulated_volume = go.Scatter(x=merged_sim_obs.index, y=sim_volume_cum, name='Simulated', )
    corrected_volume = go.Scatter(x=merged_cor_obs.index, y=corr_volume_cum, name='Corrected Simulated', )

    layout = go.Layout(
        title=build_title('Cumulative Volume Comparison', plot_titles=plot_titles),
        xaxis=dict(title='Datetime', ), yaxis=dict(title='Volume (m<sup>3</sup>)', autorange=True),
        showlegend=True)

    return go.Figure(data=[observed_volume, simulated_volume, corrected_volume], layout=layout)



import geoglows.bias
import geoglows._plots as plots
import geoglows.data
import geoglows.analyze
import geoglows.streams
import geoglows.tables
import geoglows.streamflow

from ._constants import get_metadata_table_path, set_metadata_table_path

__all__ = [
    'bias', 'plots', 'data', 'analyze', 'streams', 'tables', 'streamflow',
    'get_metadata_table_path', 'set_metadata_table_path',
]
__version__ = '1.7.0'
__author__ = 'Riley Hales'
__license__ = 'BSD 3-Clause Clear License'



import datetime

import pytz
from plotly.offline import plot as offline_plot


def build_title(main_title, plot_titles: list):
    if plot_titles is not None:
        main_title += '<br>'.join(plot_titles)
    return main_title


def return_period_plot_colors():
    return {
        '2 Year': 'rgba(254, 240, 1, .4)',
        '5 Year': 'rgba(253, 154, 1, .4)',
        '10 Year': 'rgba(255, 56, 5, .4)',
        '20 Year': 'rgba(128, 0, 246, .4)',
        '25 Year': 'rgba(255, 0, 0, .4)',
        '50 Year': 'rgba(128, 0, 106, .4)',
        '100 Year': 'rgba(128, 0, 246, .4)',
    }


def plotly_figure_to_html_plot(figure, include_plotlyjs: bool = False, ) -> str:
    return offline_plot(
        figure,
        config={'autosizable': True, 'responsive': True},
        output_type='div',
        include_plotlyjs=include_plotlyjs
    )


def timezone_label(timezone: str = None):
    timezone = str(timezone) if timezone is not None else 'UTC'
    # get the number of hours the timezone is offset from UTC
    now = datetime.datetime.now(pytz.timezone(timezone))
    utc_offset = now.utcoffset().total_seconds() / 3600
    # convert float number of hours to HH:MM format
    utc_offset = f'{int(utc_offset):+03d}:{int((utc_offset % 1) * 60):02d}'
    return f'Datetime ({timezone} {utc_offset})'



import pandas as pd
import plotly.graph_objects as go

from .format_tools import return_period_plot_colors

__all__ = [
    '_rperiod_scatters'
]


def _rperiod_scatters(startdate: str,
                      enddate: str,
                      rperiods: pd.DataFrame,
                      y_max: float,
                      max_plotted_flow: float = 0,
                      visible: bool = None):
    colors = return_period_plot_colors()
    x_vals = (startdate, enddate, enddate, startdate)

    r2 = int(rperiods[2].values[0])
    r5 = int(rperiods[5].values[0])
    r10 = int(rperiods[10].values[0])
    r25 = int(rperiods[25].values[0])
    r50 = int(rperiods[50].values[0])
    r100 = int(rperiods[100].values[0])
    rmax = int(max(2 * r100 - r25, y_max))

    visible = True if max_plotted_flow > r2 else 'legendonly'

    def template(name, y, color, fill='toself'):
        return go.Scatter(
            name=name,
            x=x_vals,
            y=y,
            legendgroup='returnperiods',
            fill=fill,
            visible=visible,
            line=dict(color=color, width=0))

    return [
        template('Return Periods', (r2, r2, rmax, rmax), 'rgba(0,0,0,0)', fill='none'),
        template(f'2 Year: {r2}', (r2, r2, r5, r5), colors['2 Year']),
        template(f'5 Year: {r5}', (r5, r5, r10, r10), colors['5 Year']),
        template(f'10 Year: {r10}', (r10, r10, r25, r25), colors['10 Year']),
        template(f'25 Year: {r25}', (r25, r25, r50, r50), colors['25 Year']),
        template(f'50 Year: {r50}', (r50, r50, r100, r100), colors['50 Year']),
        template(f'100 Year: {r100}', (r100, r100, rmax, rmax), colors['100 Year']),
    ]



import warnings
from io import StringIO

import numpy as np
import pandas as pd
import requests
import s3fs
import xarray as xr

from ._constants import (
    ODP_FORECAST_S3_BUCKET_URI,
    ODP_RETROSPECTIVE_S3_BUCKET_URI,
    ODP_S3_BUCKET_REGION,
)
from .analyze import (
    simple_forecast as calc_simple_forecast,
    forecast_stats as calc_forecast_stats,
)

DEFAULT_REST_ENDPOINT = 'https://geoglows.ecmwf.int/api/'
DEFAULT_REST_ENDPOINT_VERSION = 'v2'  # 'v1, v2, latest'

__all__ = [
    '_forecast',
    '_retrospective',
]


def _forecast(function):
    def from_aws(*args, **kwargs):
        product_name = function.__name__.replace("_", "").lower()
        if product_name == 'forecastrecords':
            warnings.warn('forecast_records are not available from the AWS Open Data Program.')
            return from_rest(*args, **kwargs)

        river_id = args[0] if len(args) > 0 else kwargs.get('river_id', None)
        if river_id is None and product_name != 'dates':
            raise ValueError('River ID must be provided to retrieve forecast data.')

        return_format = kwargs.get('format', 'df')
        assert return_format in ('df', 'xarray'), f'Unsupported return format requested: {return_format}'

        if kwargs.get('skip_log', False):
            requests.post(f'{DEFAULT_REST_ENDPOINT}{DEFAULT_REST_ENDPOINT_VERSION}/log',
                          json={'river_id': river_id, 'product': product_name, 'format': return_format},
                          timeout=1, )  # short timeout- don't need the response, post only needs to be received

        s3 = s3fs.S3FileSystem(anon=True, client_kwargs=dict(region_name=ODP_S3_BUCKET_REGION))
        date = kwargs.get('date', False)
        if not date:
            zarr_vars = ['rivid', 'Qout', 'time', 'ensemble']
            dates = [s3.glob(ODP_FORECAST_S3_BUCKET_URI + '/' + f'*.zarr/{var}') for var in zarr_vars]
            dates = [set([d.split('/')[1].replace('.zarr', '') for d in date]) for date in dates]
            dates = sorted(set.intersection(*dates), reverse=True)
            if product_name == 'dates':
                return pd.DataFrame(dict(dates=dates))
            date = dates[0]
        if len(date) == 8:
            date = f'{date}00.zarr'
        elif len(date) == 10:
            date = f'{date}.zarr'
        else:
            raise ValueError('Date must be YYYYMMDD or YYYYMMDDHH format. Use dates() to view available data.')

        s3store = s3fs.S3Map(root=f'{ODP_FORECAST_S3_BUCKET_URI}/{date}', s3=s3, check=False)

        attrs = {
            'source': 'geoglows',
            'forecast_date': date[:8],
            'retrieval_date': pd.Timestamp.now().strftime('%Y%m%d'),
            'units': 'cubic meters per second',
        }
        ds = xr.open_zarr(s3store).sel(rivid=river_id)
        if return_format == 'xarray' and product_name == 'forecastensembles':
            ds = ds.rename({'time': 'datetime', 'rivid': 'river_id'})
            ds.attrs = attrs
            return ds
        df = ds.to_dataframe().round(2).reset_index()
        df['time'] = pd.to_datetime(df['time'], utc=True)

        # rename columns to match the REST API
        if isinstance(river_id, int) or isinstance(river_id, np.int64):
            df = df.pivot(index='time', columns='ensemble', values='Qout')
        else:
            df = df.pivot(index=['time', 'rivid'], columns='ensemble', values='Qout')
            df.index.names = ['time', 'river_id']
        df = df[sorted(df.columns)]
        df.columns = [f'ensemble_{str(x).zfill(2)}' for x in df.columns]

        if product_name == 'forecast':
            df = calc_simple_forecast(df)
        elif product_name == 'forecaststats':
            df = calc_forecast_stats(df)

        if return_format == 'df':
            return df
        ds = df.to_xarray()
        ds.attrs = attrs
        return ds

    def from_rest(*args, **kwargs):
        # update the default values set by the function unless the user has already specified them
        for key, value in function.__kwdefaults__.items() if function.__kwdefaults__ else []:
            if key not in kwargs:
                kwargs[key] = value

        return_format = kwargs.get('format', 'csv')
        assert return_format in ('csv', 'json', 'url'), f'Unsupported format requested: {return_format}'

        # parse out the information necessary to build a request url
        endpoint = kwargs.get('endpoint', DEFAULT_REST_ENDPOINT)
        endpoint = endpoint[:-1] if endpoint[-1] == '/' else endpoint
        endpoint = endpoint + '/api' if not endpoint.endswith('/api') else endpoint
        endpoint = f'https://{endpoint}' if not endpoint.startswith(('https://', 'http://')) else endpoint

        version = kwargs.get('version', DEFAULT_REST_ENDPOINT_VERSION)
        assert version in ('v2',), ValueError(f'Unrecognized model version parameter: {version}')

        product_name = function.__name__.replace("_", "").lower()

        river_id = args[0] if len(args) > 0 else kwargs.get('river_id', None)
        if river_id is None and product_name != 'dates':
            raise ValueError('River ID must be provided to retrieve forecast data.')
        if not isinstance(river_id, (int, np.int64, )):
            raise ValueError('Multiple river_ids are not available via REST API. Provide a single 9 digit integer.')
        if river_id and version == 'v2':
            assert 1_000_000_000 > river_id >= 110_000_000, ValueError('River ID must be a 9 digit integer')

        # request parameter validation before submitting
        for key in ('endpoint', 'version', 'river_id'):
            if key in kwargs:
                del kwargs[key]
        for key, value in kwargs.items():
            if value is None:
                del kwargs[key]
        for date in ('date', 'start_date', 'end_date'):
            if date in kwargs:
                assert len(str(kwargs[date])) == 8 or len(
                    str(kwargs[date])) == 10, f'Invalid date format: {kwargs[date]}'
        if 'format' in kwargs and kwargs['format'] != 'json':
            del kwargs['format']
        kwargs['source'] = kwargs.get('source', 'pygeoglows')  # allow using default for specific apps which override
        params = '&'.join([f'{key}={value}' for key, value in kwargs.items()])

        # piece together the request url
        request_url = f'{endpoint}/{version}/{product_name}'  # build the base url
        request_url = f'{request_url}/{river_id}' if river_id else request_url  # add the river_id if it exists
        request_url = f'{request_url}?{params}'  # add the query parameters

        if return_format == 'url':
            return request_url.replace(f'source={kwargs["source"]}', '')

        response = requests.get(request_url)

        if response.status_code != 200:
            raise RuntimeError('Received an error from the REST API: ' + response.text)

        if return_format == 'csv':
            df = pd.read_csv(StringIO(response.text))
            if 'datetime' in df.columns:
                df['datetime'] = pd.to_datetime(df['datetime'])
                df = df.set_index('datetime')
                df.index = df.index.tz_localize('UTC')
            return df
        elif return_format == 'json':
            return response.json()
        else:
            raise ValueError(f'Unsupported return format requested: {return_format}')

    def main(*args, **kwargs):
        source = kwargs.get('data_source', 'aws')
        assert source in ('rest', 'aws'), ValueError(f'Unrecognized data source requested: {source}')
        if source == 'rest':
            return from_rest(*args, **kwargs)
        return from_aws(*args, **kwargs)

    main.__doc__ = function.__doc__  # necessary for code documentation auto generators
    return main


def _retrospective(function):
    def main(*args, **kwargs):
        product_name = function.__name__.replace("_", "-").lower()

        river_id = args[0] if len(args) > 0 else kwargs.get('river_id', None)
        if river_id is None:
            raise ValueError('River ID must be provided to retrieve retrospective data.')

        return_format = kwargs.get('format', 'df')
        assert return_format in ('df', 'xarray'), f'Unsupported return format requested: {return_format}'

        method = kwargs.get('method', 'gumbel1')

        if kwargs.get('skip_log', False):
            requests.post(f'{DEFAULT_REST_ENDPOINT}{DEFAULT_REST_ENDPOINT_VERSION}/log',
                          timeout=1,  # short timeout because we don't need the response, post just needs to be received
                          json={'river_id': river_id, 'product': product_name, 'format': return_format})

        s3 = s3fs.S3FileSystem(anon=True, client_kwargs=dict(region_name=ODP_S3_BUCKET_REGION))
        s3store = s3fs.S3Map(root=f'{ODP_RETROSPECTIVE_S3_BUCKET_URI}/{product_name}.zarr', s3=s3, check=False)
        ds = xr.open_zarr(s3store)
        try:
            ds = ds.sel(rivid=river_id)
        except Exception:
            raise ValueError(f'River ID(s) not found in the retrospective dataset: {river_id}')
        if return_format == 'xarray':
            return ds
        if product_name == 'retrospective':
            df = (
                ds
                .to_dataframe()
                .reset_index()
                .set_index('time')
                .pivot(columns='rivid', values='Qout')
            )
            df.index = df.index.tz_localize('UTC')
            return df
        if product_name == 'return-periods':
            rp_methods = {
                'gumbel1': 'gumbel1_return_period',
            }
            assert method in rp_methods, f'Unrecognized return period estimation method given: {method}'
            return (
                ds
                [rp_methods[method]]
                .to_dataframe()
                .reset_index()
                .pivot(index='rivid', columns='return_period', values=rp_methods[method])
            )
        raise ValueError(f'Unsupported product requested: {product_name}')

    main.__doc__ = function.__doc__  # necessary for code documentation auto generators
    return main



import numpy as np
import pandas as pd
import plotly.graph_objects as go

from .format_tools import build_title, timezone_label
from .plotly_helpers import _rperiod_scatters

__all__ = [
    'forecast',
    'forecast_stats',
    'forecast_ensembles',
    'forecast_records',
]


def forecast(df: pd.DataFrame, *,
             rp_df: pd.DataFrame = None,
             plot_titles: list = None, ) -> go.Figure:
    """
    Plots forecasted streamflow and optional return periods
    Args:
        df: the dataframe response from geoglows.data.forecast
        rp_df: optional dataframe of return period data
        plot_titles: optional list of strings to include in the figure title. each list item will be on a new line.

    Returns:
        go.Figure
    """
    scatter_traces = [
        go.Scatter(
            x=df.index,
            y=df['flow_median'],
            name='Streamflow (Median)',
            line=dict(color='black'),
        ),
        go.Scatter(
            name='Uncertainty Bounds',
            x=np.concatenate([df.index, df.index[::-1]]),
            y=np.concatenate([df['flow_uncertainty_upper'], df['flow_uncertainty_lower'][::-1]]),
            legendgroup='uncertainty',
            showlegend=True,
            fill='toself',
            line=dict(color='lightblue', dash=None)
        ),
        go.Scatter(
            name='Uncertainty Upper Bounds (80%)',
            x=df.index,
            y=df['flow_uncertainty_upper'],
            legendgroup='uncertainty',
            showlegend=False,
            line=dict(color='lightblue', dash='dash')
        ),
        go.Scatter(
            name='Uncertainty Lower Bounds (20%)',
            x=df.index,
            y=df['flow_uncertainty_lower'],
            legendgroup='uncertainty',
            showlegend=False,
            line=dict(color='lightblue', dash='dash')
        ),
    ]

    if rp_df is not None:
        rperiod_scatters = _rperiod_scatters(df.index[0], df.index[-1], rp_df, df['flow_uncertainty_upper'].max())
        scatter_traces += rperiod_scatters

    layout = go.Layout(
        title=build_title('Forecasted Streamflow', plot_titles),
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)', 'range': [0, 'auto']},
        xaxis={
            'title': timezone_label(df.index.tz),
            'range': [df.index[0], df.index[-1]],
            'hoverformat': '%d %b %Y %X',
        },
    )

    return go.Figure(scatter_traces, layout=layout)


def forecast_stats(df: pd.DataFrame, *,
                   rp_df: pd.DataFrame = None,
                   plot_titles: list = None,
                   show_maxmin: bool = False, ) -> go.Figure:
    """
    Makes the streamflow data and optional metadata into a plotly plot

    Args:
        df: the csv response from forecast_stats
        rp_df: the csv response from return_periods
        plot_titles: a list of strings to place in the figure title. each list item will be on a new line.
        show_maxmin: Choose to show or hide the max/min envelope by default

    Return:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    # Start processing the inputs
    dates = df.index.tolist()
    startdate = dates[0]
    enddate = dates[-1]

    plot_data = {
        'x_stats': df['flow_avg'].dropna(axis=0).index.tolist(),
        'x_hires': df['high_res'].dropna(axis=0).index.tolist(),
        'y_max': max(df['flow_max']),
        'flow_max': list(df['flow_max'].dropna(axis=0)),
        'flow_75%': list(df['flow_75p'].dropna(axis=0)),
        'flow_avg': list(df['flow_avg'].dropna(axis=0)),
        'flow_med': list(df['flow_med'].dropna(axis=0)),
        'flow_25%': list(df['flow_25p'].dropna(axis=0)),
        'flow_min': list(df['flow_min'].dropna(axis=0)),
        'high_res': list(df['high_res'].dropna(axis=0)),
    }

    maxmin_visible = True if show_maxmin else 'legendonly'
    scatter_plots = [
        # Plot together so you can use fill='toself' for the shaded box, also separately so the labels appear
        go.Scatter(name='Maximum & Minimum Flow',
                   x=plot_data['x_stats'] + plot_data['x_stats'][::-1],
                   y=plot_data['flow_max'] + plot_data['flow_min'][::-1],
                   legendgroup='boundaries',
                   fill='toself',
                   visible=maxmin_visible,
                   line=dict(color='lightblue', dash='dash')),
        go.Scatter(name='Maximum',
                   x=plot_data['x_stats'],
                   y=plot_data['flow_max'],
                   legendgroup='boundaries',
                   visible=maxmin_visible,
                   showlegend=False,
                   line=dict(color='darkblue', dash='dash')),
        go.Scatter(name='Minimum',
                   x=plot_data['x_stats'],
                   y=plot_data['flow_min'],
                   legendgroup='boundaries',
                   visible=maxmin_visible,
                   showlegend=False,
                   line=dict(color='darkblue', dash='dash')),

        go.Scatter(name='25-75 Percentile Flow',
                   x=plot_data['x_stats'] + plot_data['x_stats'][::-1],
                   y=plot_data['flow_75%'] + plot_data['flow_25%'][::-1],
                   legendgroup='percentile_flow',
                   fill='toself',
                   line=dict(color='lightgreen'), ),
        go.Scatter(name='75%',
                   x=plot_data['x_stats'],
                   y=plot_data['flow_75%'],
                   showlegend=False,
                   legendgroup='percentile_flow',
                   line=dict(color='green'), ),
        go.Scatter(name='25%',
                   x=plot_data['x_stats'],
                   y=plot_data['flow_25%'],
                   showlegend=False,
                   legendgroup='percentile_flow',
                   line=dict(color='green'), ),

        go.Scatter(name='Higher Time Step Forecast',
                   x=plot_data['x_hires'],
                   y=plot_data['high_res'],
                   visible=True,
                   line={'color': 'black'}, ),

        go.Scatter(name='Average Flow',
                   x=plot_data['x_stats'],
                   y=plot_data['flow_avg'],
                   line=dict(color='blue'), ),
        go.Scatter(name='Median Flow',
                   x=plot_data['x_stats'],
                   y=plot_data['flow_med'],
                   line=dict(color='red'), ),
    ]
    if rp_df is not None:
        plot_data.update(rp_df.to_dict(orient='index').items())
        max_visible = max(max(plot_data['flow_75%']), max(plot_data['flow_avg']), max(plot_data['high_res']))
        rperiod_scatters = _rperiod_scatters(startdate, enddate, rp_df, plot_data['y_max'], max_visible)
        scatter_plots += rperiod_scatters

    layout = go.Layout(
        title=build_title('Forecasted Streamflow', plot_titles),
        yaxis={
            'title': 'Streamflow (m<sup>3</sup>/s)',
            'range': [0, 'auto']
        },
        xaxis={
            'title': timezone_label(df.index.tz),
            'range': [startdate, enddate],
            'hoverformat': '%d %b %Y %X',
            'tickformat': '%b %d %Y'
        },
    )

    return go.Figure(scatter_plots, layout=layout)


def forecast_ensembles(df: pd.DataFrame, *, rp_df: pd.DataFrame = None, plot_titles: list = None, ) -> go.Figure:
    """
    Makes the streamflow ensemble data and metadata into a plotly plot

    Args:
        df: the dataframe response from geoglows.data.forecast_ensembles
        rp_df: the csv response from return_periods
        plot_titles: a list of strings to place in the figure title. each list item will be on a new line.

    Return:
         go.Figure
    """
    # variables to determine the maximum flow and hold all the scatter plot lines
    max_flows = []
    scatter_plots = []

    # determine the threshold values for return periods and the start/end dates so we can plot them
    dates = df.index.tolist()
    startdate = dates[0]
    enddate = dates[-1]

    # process the series' components and store them in a dictionary
    plot_data = {
        'x_1-51': df['ensemble_01'].dropna(axis=0).index.tolist(),
        'x_52': df['ensemble_52'].dropna(axis=0).index.tolist(),
    }

    # add a dictionary entry for each of the ensemble members. the key for each series is the integer ensemble number
    for ensemble in df.columns:
        plot_data[ensemble] = df[ensemble].dropna(axis=0).tolist()
        max_flows.append(max(plot_data[ensemble]))
    plot_data['y_max'] = max(max_flows)

    if rp_df is not None:
        plot_data.update(rp_df.to_dict(orient='index').items())
        rperiod_scatters = _rperiod_scatters(startdate, enddate, rp_df, plot_data['y_max'])
    else:
        rperiod_scatters = []

    # create the high resolution line (ensemble 52)
    scatter_plots.append(go.Scatter(
        name='High Resolution Forecast',
        x=plot_data['x_52'],
        y=plot_data['ensemble_52'],
        line=dict(color='black')
    ))
    # create a line for the rest of the ensembles (1-51)
    for i in range(1, 52):
        scatter_plots.append(go.Scatter(
            name='Ensemble ' + str(i),
            x=plot_data['x_1-51'],
            y=plot_data[f'ensemble_{i:02}'],
            legendgroup='Ensemble Members'
        ))
    scatter_plots += rperiod_scatters

    # define a layout for the plot
    layout = go.Layout(
        title=build_title('Ensemble Predicted Streamflow', plot_titles),
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)', 'range': [0, 'auto']},
        xaxis={
            'title': timezone_label(df.index.tz),
            'range': [startdate, enddate],
            'hoverformat': '%d %b %Y %X',
            'tickformat': '%b %d %Y'
        },
    )
    return go.Figure(scatter_plots, layout=layout)


def forecast_records(df: pd.DataFrame, *, rp_df: pd.DataFrame = None, plot_titles: list = False, ) -> go.Figure:
    """
    Makes the streamflow saved forecast data and metadata into a plotly plot

    Args:
        df: the csv response from forecast_records
        rp_df: the csv response from return_periods
        plot_titles: a list of strings to place in the figure title. each list item will be on a new line.

    Return:
         plotly.GraphObject: plotly object, especially for use with python notebooks and the .show() method
    """
    # Start processing the inputs
    dates = df.index.tolist()
    startdate = dates[0]
    enddate = dates[-1]

    plot_data = {
        'x_records': dates,
        'recorded_flows': df.dropna(axis=0).values.flatten(),
        'y_max': np.nanmax(df.values),
    }
    if rp_df is not None:
        plot_data.update(rp_df.to_dict(orient='index').items())
        rperiod_scatters = _rperiod_scatters(startdate, enddate, rp_df, plot_data['y_max'], plot_data['y_max'])
    else:
        rperiod_scatters = []

    scatter_plots = [go.Scatter(
        name='Previous Forecast Average',
        x=plot_data['x_records'],
        y=plot_data['recorded_flows'],
        line=dict(color='gold'),
    )] + rperiod_scatters

    layout = go.Layout(
        title=build_title('Previous Forecasted Streamflow', plot_titles=plot_titles),
        yaxis={'title': 'Streamflow (m<sup>3</sup>/s)', 'range': [0, 'auto']},
        xaxis={
            'title': timezone_label(df.index.tz),
            'range': [startdate, enddate],
            'hoverformat': '%d %b %Y %X',
        },
    )
    return go.Figure(scatter_plots, layout=layout)



import math
import warnings

import hydrostats as hs
import hydrostats.data as hd
import numpy as np
import pandas as pd
from scipy import interpolate

__all__ = [
    'correct_historical',
    'correct_forecast',
    'statistics_tables',
]


def correct_historical(simulated_data: pd.DataFrame, observed_data: pd.DataFrame) -> pd.DataFrame:
    """
    Accepts a historically simulated flow timeseries and observed flow timeseries and attempts to correct biases in the
    simulation on a monthly basis.

    Args:
        simulated_data: A dataframe with a datetime index and a single column of streamflow values
        observed_data: A dataframe with a datetime index and a single column of streamflow values

    Returns:
        pandas DataFrame with a datetime index and a single column of streamflow values
    """
    # list of the unique months in the historical simulation. should always be 1->12 but just in case...
    unique_simulation_months = sorted(set(simulated_data.index.strftime('%m')))
    dates = []
    values = []

    for month in unique_simulation_months:
        # filter historic data to only be current month
        monthly_simulated = simulated_data[simulated_data.index.month == int(month)].dropna()
        to_prob = _flow_and_probability_mapper(monthly_simulated, to_probability=True)
        # filter the observations to current month
        monthly_observed = observed_data[observed_data.index.month == int(month)].dropna()
        to_flow = _flow_and_probability_mapper(monthly_observed, to_flow=True)

        dates += monthly_simulated.index.to_list()
        value = to_flow(to_prob(monthly_simulated.values))
        values += value.tolist()

    corrected = pd.DataFrame(data=values, index=dates, columns=['Corrected Simulated Streamflow'])
    corrected.sort_index(inplace=True)
    return corrected


def correct_forecast(forecast_data: pd.DataFrame, simulated_data: pd.DataFrame,
                     observed_data: pd.DataFrame, use_month: int = 0) -> pd.DataFrame:
    """
    Accepts a short term forecast of streamflow, simulated historical flow, and observed flow timeseries and attempts
    to correct biases in the forecasted data

    Args:
        forecast_data: A dataframe with a datetime index and any number of columns of forecasted flow. Compatible with
            forecast_stats, forecast_ensembles, forecast_records
        simulated_data: A dataframe with a datetime index and a single column of streamflow values
        observed_data: A dataframe with a datetime index and a single column of streamflow values
        use_month: Optional: either 0 for correct the forecast based on the first month of the forecast data or -1 if
            you want to correct based on the ending month of the forecast data

    Returns:
        pandas DataFrame with a copy of forecasted data with values updated in each column
    """
    # make a copy of the forecasts which we update and return so the original data is not changed
    forecast_copy = forecast_data.copy()

    # make the flow and probability interpolation functions
    monthly_simulated = simulated_data[simulated_data.index.month == forecast_copy.index[use_month].month].dropna()
    monthly_observed = observed_data[observed_data.index.month == forecast_copy.index[use_month].month].dropna()
    to_prob = _flow_and_probability_mapper(monthly_simulated, to_probability=True, extrapolate=True)
    to_flow = _flow_and_probability_mapper(monthly_observed, to_flow=True, extrapolate=True)

    # for each column of forecast data, make the interpolation function and update the dataframe
    for column in forecast_copy.columns:
        tmp = forecast_copy[column].dropna()
        forecast_copy.update(pd.DataFrame(to_flow(to_prob(tmp.values)), index=tmp.index, columns=[column]))

    return forecast_copy


def statistics_tables(corrected: pd.DataFrame, simulated: pd.DataFrame, observed: pd.DataFrame,
                      merged_sim_obs: pd.DataFrame = False, merged_cor_obs: pd.DataFrame = False,
                      metrics: list = None) -> str:
    """
    Makes an html table of various statistical metrics for corrected vs observed data alongside the same metrics for
    the simulated vs observed data as a way to see the improvement made by the bias correction. This function uses
    hydrostats.data.merge_data on the 3 inputs. If you have already computed these because you are doing a full
    comparison of bias correction, you can provide them to save time

    Args:
        corrected: A dataframe with a datetime index and a single column of streamflow values
        simulated: A dataframe with a datetime index and a single column of streamflow values
        observed: A dataframe with a datetime index and a single column of streamflow values
        merged_sim_obs: (optional) if you have already computed it, hydrostats.data.merge_data(simulated, observed)
        merged_cor_obs: (optional) if you have already computed it, hydrostats.data.merge_data(corrected, observed)
        metrics: A list of abbreviated statistic names. See the documentation for HydroErr
    """
    if corrected is False and simulated is False and observed is False:
        if merged_sim_obs is not False and merged_cor_obs is not False:
            pass  # if you provided the merged dataframes already, we use those
    else:
        # merge the datasets together
        merged_sim_obs = hd.merge_data(sim_df=simulated, obs_df=observed)
        merged_cor_obs = hd.merge_data(sim_df=corrected, obs_df=observed)

    if metrics is None:
        metrics = ['ME', 'RMSE', 'NRMSE (Mean)', 'MAPE', 'NSE', 'KGE (2009)', 'KGE (2012)']
    # Merge Data
    table1 = hs.make_table(merged_dataframe=merged_sim_obs, metrics=metrics)
    table2 = hs.make_table(merged_dataframe=merged_cor_obs, metrics=metrics)

    table2 = table2.rename(index={'Full Time Series': 'Corrected Full Time Series'})
    table1 = table1.rename(index={'Full Time Series': 'Original Full Time Series'})
    table1 = table1.transpose()
    table2 = table2.transpose()

    table_final = pd.merge(table1, table2, right_index=True, left_index=True)

    return table_final.to_html()


def _flow_and_probability_mapper(monthly_data: pd.DataFrame, to_probability: bool = False,
                                 to_flow: bool = False, extrapolate: bool = False) -> interpolate.interp1d:
    if not to_flow and not to_probability:
        raise ValueError('You need to specify either to_probability or to_flow as True')

    # get maximum value to bound histogram
    max_val = math.ceil(np.max(monthly_data.max()))
    min_val = math.floor(np.min(monthly_data.min()))

    if max_val == min_val:
        warnings.warn('The observational data has the same max and min value. You may get unanticipated results.')
        max_val += .1

    # determine number of histograms bins needed
    number_of_points = len(monthly_data.values)
    number_of_classes = math.ceil(1 + (3.322 * math.log10(number_of_points)))

    # specify the bin width for histogram (in m3/s)
    step_width = (max_val - min_val) / number_of_classes

    # specify histogram bins
    bins = np.arange(-np.min(step_width), max_val + 2 * np.min(step_width), np.min(step_width))

    if bins[0] == 0:
        bins = np.concatenate((-bins[1], bins))
    elif bins[0] > 0:
        bins = np.concatenate((-bins[0], bins))

    # make the histogram
    counts, bin_edges = np.histogram(monthly_data, bins=bins)

    # adjust the bins to be the center
    bin_edges = bin_edges[1:]

    # normalize the histograms
    counts = counts.astype(float) / monthly_data.size

    # calculate the cdfs
    cdf = np.cumsum(counts)

    # interpolated function to convert simulated streamflow to prob
    if to_probability:
        if extrapolate:
            func = interpolate.interp1d(bin_edges, cdf, fill_value='extrapolate')
        else:
            func = interpolate.interp1d(bin_edges, cdf)
        return lambda x: np.clip(func(x), 0, 1)
    # interpolated function to convert simulated prob to observed streamflow
    elif to_flow:
        if extrapolate:
            return interpolate.interp1d(cdf, bin_edges, fill_value='extrapolate')
        return interpolate.interp1d(cdf, bin_edges)



import pandas as pd

__all__ = [
    'flood_probabilities',
    'return_periods',
]


def flood_probabilities(ensem: pd.DataFrame, rperiods: pd.DataFrame) -> str:
    """
    Processes the results of forecast_ensembles and return_periods and shows the probabilities of exceeding the
    return period flow on each day.

    Args:
        ensem: the csv response from forecast_ensembles
        rperiods: the csv response from return_periods

    Return:
         string containing html to build a table with a row of dates and for exceeding each return period threshold
    """
    ens = ensem.drop(columns=['ensemble_52']).dropna()
    ens = ens.groupby(ens.index.date).max()
    ens.index = pd.to_datetime(ens.index).strftime('%Y-%m-%d')
    # for each return period, get the percentage of columns with a value greater than the return period on each day
    percent_series = {rp: (ens > rperiods[rp].values[0]).mean(axis=1).values.tolist() for rp in rperiods}
    percent_series = pd.DataFrame(percent_series, index=ens.index)
    percent_series.index.name = 'Date'
    percent_series.columns = [f'{c} Year' for c in percent_series.columns]
    percent_series = percent_series * 100
    percent_series = percent_series.round(1)
    percent_series = percent_series.reset_index()

    colors = {
        '2 Year': 'rgba(254, 240, 1, {0})',
        '5 Year': 'rgba(253, 154, 1, {0})',
        '10 Year': 'rgba(255, 56, 5, {0})',
        '20 Year': 'rgba(128, 0, 246, {0})',
        '25 Year': 'rgba(255, 0, 0, {0})',
        '50 Year': 'rgba(128, 0, 106, {0})',
        '100 Year': 'rgba(128, 0, 246, {0})',
    }

    headers = "".join([f'<th>{x}</th>' for x in percent_series.columns])
    rows = []
    for row_idx, row in enumerate(percent_series.values.tolist()):
        cells = []
        for col_idx, cell in enumerate(row):
            if col_idx == 0:
                cells.append(f'<td>{cell}</td>')
                continue
            column_name = percent_series.columns[col_idx]
            stylestring = f'background-color: {colors[column_name].format(round(cell * 0.005, 2))}'
            cells.append(f'<td style="{stylestring}">{cell}</td>')
        rows.append(f'<tr>{"".join(cells)}</tr>')
    rows = "".join(rows)

    return f'<table id="probabilities_table" align="center"><tbody><tr>{headers}</tr>{rows}</tbody></table>'


def return_periods(rperiods: pd.DataFrame) -> str:
    """
    Makes an html string of a table of return periods and river id numbers

    Args:
        rperiods: the dataframe from the return periods function

    Returns:
        string of html
    """
    colors = {
        '2 Year': 'rgba(254, 240, 1, .4)',
        '5 Year': 'rgba(253, 154, 1, .4)',
        '10 Year': 'rgba(255, 56, 5, .4)',
        '20 Year': 'rgba(128, 0, 246, .4)',
        '25 Year': 'rgba(255, 0, 0, .4)',
        '50 Year': 'rgba(128, 0, 106, .4)',
        '100 Year': 'rgba(128, 0, 246, .4)',
    }
    rperiods = rperiods.astype(float).round(2).astype(str)
    rperiods.columns = [f'{c} Year' for c in rperiods.columns]
    rperiods.index = rperiods.index.astype(int).astype(str)
    rperiods = rperiods.reset_index()

    headers = "".join([f'<th>{x}</th>' for x in rperiods.columns])
    rows = []
    for row in rperiods.values.tolist():
        cells = []
        for col_idx, cell in enumerate(row):
            if col_idx == 0:
                cells.append(f'<td>{cell}</td>')
                continue
            column_name = rperiods.columns[col_idx]
            stylestring = f'background-color: {colors[column_name]}'
            cells.append(f'<td style="{stylestring}">{cell}</td>')
        rows.append(f'<tr>{"".join(cells)}</tr>')
    rows = "".join(rows)

    return f'<table id="return_periods_table" align="center"><tbody><tr>{headers}</tr>{rows}</tbody></table>'



import re

from setuptools import setup, find_packages

NAME = 'geoglows'
DESCRIPTION = 'Package for accessing data from the GEOGLOWS Hydrological Model'
URL = 'https://data.geoglows.org'
AUTHOR = 'Riley Hales PhD'
REQUIRES_PYTHON = '>=3.7.0'
LICENSE = 'BSD 3-Clause Clear License'

with open("README.md", "r") as readme:
    LONG_DESCRIPTION = readme.read()

with open(f'./{NAME}/__init__.py') as f:
    version_pattern = r'__version__ = [\'"](.+)[\'"]'
    VERSION = re.search(version_pattern, f.read()).group(1)

with open('./requirements.txt') as f:
    INSTALL_REQUIRES = f.read().splitlines()

setup(
    name=NAME,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    url=URL,
    project_urls=dict(
        Homepage='https://data.geoglows.org',
        Documentation='https://geoglows.readthedocs.io',
        Source='https://github.com/geoglows/pygeoglows',
    ),
    license=LICENSE,
    license_family='BSD',
    package_data={'': ['*.ipynb', '*.html']},
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Hydrology',
        'Topic :: Scientific/Engineering :: Visualization',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
    ],
    install_requires=INSTALL_REQUIRES,
    python_requires=REQUIRES_PYTHON,
)



import os

ODP_CORE_S3_BUCKET_URI = 's3://geoglows-v2'
ODP_FORECAST_S3_BUCKET_URI = 's3://geoglows-v2-forecasts'
ODP_RETROSPECTIVE_S3_BUCKET_URI = 's3://geoglows-v2-retrospective'
ODP_S3_BUCKET_REGION = 'us-west-2'

METADATA_TABLE_PATH = os.getenv(
    'PYGEOGLOWS_METADATA_TABLE_PATH',
    os.path.join(os.path.dirname(__file__), 'data', 'metadata-tables.parquet')
)


def get_metadata_table_path() -> str:
    return METADATA_TABLE_PATH


def set_metadata_table_path(path: str) -> str:
    global METADATA_TABLE_PATH
    METADATA_TABLE_PATH = path
    return METADATA_TABLE_PATH



import math

import numpy as np
import pandas as pd

__all__ = [
    'gumbel1',
    'simple_forecast',
    'forecast_stats',
    'daily_averages',
    'monthly_averages',
    'annual_averages',
    'daily_stats',
    'daily_variance',
    'daily_flow_anomaly',
    'return_periods',
    'low_return_periods'
]


def gumbel1(rp: int, xbar: float, std: float) -> float:
    """
    Solves the Gumbel Type 1 distribution
    Args:
        rp: return period (years)
        xbar: average of the dataset
        std: standard deviation of the dataset

    Returns:
        float: solution to gumbel distribution
    """
    return round(-math.log(-math.log(1 - (1 / rp))) * std * .7797 + xbar - (.45 * std), 2)


def simple_forecast(ens: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates the simple forecast from a dataframe of forecast ensembles

    Args:
        ens: a dataframe of forecast ensembles

    Returns:
        pandas DataFrame with datetime index and columns flow_uncertainty_upper, flow_median, flow_uncertainty_lower
    """
    df = (
        ens
        .drop(columns=['ensemble_52'])
        .dropna()
    )
    df = pd.DataFrame({
        f'flow_uncertainty_upper': np.nanpercentile(df.values, 80, axis=1),
        f'flow_median': np.median(df.values, axis=1),
        f'flow_uncertainty_lower': np.nanpercentile(df.values, 20, axis=1),
    }, index=df.index)
    return df


def forecast_stats(ens: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates the statistics for a dataframe of forecast ensembles

    Args:
        ens: a dataframe of forecast ensembles

    Returns:
        pandas DataFrame with an index of datetime and columns min, max, mean, median, 25%, 75%
    """
    df = (
        ens
        .drop(columns=['ensemble_52'])
        .dropna()
    )
    return (
        pd
        .DataFrame(
            {
                'flow_min': df.min(axis=1),
                'flow_25p': df.quantile(.25, axis=1),
                'flow_avg': df.mean(axis=1),
                'flow_med': df.median(axis=1),
                'flow_75p': df.quantile(.75, axis=1),
                'flow_max': df.max(axis=1),
            },
            index=df.index
        )
        .merge(
            ens[['ensemble_52']]
            .rename(columns={'ensemble_52': 'high_res'}),
            left_index=True,
            right_index=True,
            how='outer'
        )
        .sort_index(ascending=True)
    )


def daily_averages(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates the daily average of a dataframe with a datetime index

    Args:
        df: a dataframe of retrospective simulation data

    Returns:
        pandas DataFrame with an index of "%m/%d" dates for each day of the year
    """
    return df.groupby(df.index.strftime('%m/%d')).mean()


def monthly_averages(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates the monthly average of a dataframe with a datetime index

    Args:
        df: a dataframe of retrospective simulation data

    Returns:
        pandas DataFrame with an index of "%m" values for each month
    """
    return df.groupby(df.index.strftime('%m')).mean()


def annual_averages(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates the annual average of a dataframe with a datetime index

    Args:
        df: a dataframe of retrospective simulation data

    Returns:
        pandas DataFrame with an index of "%Y" values for each year
    """
    # select years with >= 365 entries
    df = df.groupby(df.index.strftime('%Y')).filter(lambda x: len(x) >= 365)
    return df.groupby(df.index.strftime('%Y')).mean()


def daily_variance(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculate the daily standard deviation of a dataframe with a datetime index

    Args:
        df: a dataframe of retrospective simulation data

    Returns:
        pandas DataFrame with an index of "%m/%d" dates for each day of the year
    """
    return df.groupby(df.index.strftime('%m/%d')).std()


def daily_stats(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculates the statistics for a datafame given the day of year

    Args:
      df: historical data to compute daily statistics from

    Return:
      pd.DataFrame: dataframe with average, min, 25% values, median, 75% value and max value for each day of year
    """
    daily_grouped = df.groupby(df.index.strftime('%m/%d'))
    return (
        daily_grouped
        .mean()
        .merge(daily_grouped.min(), left_index=True, right_index=True, suffixes=('_avg', '_min'))
        .merge(daily_grouped.quantile(.25), left_index=True, right_index=True)
        .merge(daily_grouped.median(), left_index=True, right_index=True, suffixes=('_25%', '_med'))
        .merge(daily_grouped.quantile(.75), left_index=True, right_index=True)
        .merge(daily_grouped.max(), left_index=True, right_index=True, suffixes=('_75%', '_max'))
    )


def daily_flow_anomaly(stats: pd.DataFrame, day_avgs: pd.DataFrame, daily: bool = True) -> pd.DataFrame:
    """
    Compute the anomaly between the average forecasted flow and the daily average

    Args:
        stats: the csv response from the ForecastStats data service
        day_avgs: the csv response from the DailyAverages data service
        daily: if true, aggregate the hourly forecast to a daily average before computing the anomaly

    Returns:
        pandas DataFrame with a datetime index and a column labeled 'anomaly_m^3/s'
    """
    anomaly_df = pd.DataFrame(stats['flow_avg_m^3/s'], index=stats.index)
    anomaly_df = anomaly_df.dropna()

    if daily:
        anomaly_df = anomaly_df.resample('D').mean()

    anomaly_df['datetime'] = anomaly_df.index
    anomaly_df.index = anomaly_df.index.strftime("%m/%d")
    anomaly_df = anomaly_df.join(day_avgs, how="inner")
    anomaly_df['anomaly_m^3/s'] = anomaly_df['flow_avg_m^3/s'] - anomaly_df['streamflow_m^3/s']
    anomaly_df.index = anomaly_df['datetime']
    del anomaly_df['flow_avg_m^3/s'], anomaly_df['datetime'], anomaly_df['streamflow_m^3/s']

    return anomaly_df


def return_periods(df: pd.DataFrame, rps: int or tuple = (2, 5, 10, 25, 50, 100)) -> dict:
    """
    Solves the Gumbel Type-I distribution using the annual maximum flow from the historic simulation

    Args:
        df: a dataframe of retrospective simulation data
        rps: an integer or iterable of integer return period numbers to compute

    Returns:
        dict with keys 'max_simulated' and 'return_period_{year}' for each year with float values to 2 decimals
    """
    annual_max_flow_list = df.groupby(df.index.strftime('%Y')).max().values
    xbar = np.mean(annual_max_flow_list)
    std = np.std(annual_max_flow_list)

    if type(rps) is int:
        rps = (rps,)

    ret_pers = {
        'max_simulated': round(np.max(annual_max_flow_list), 2)
    }
    ret_pers.update({f'return_period_{rp}': round(gumbel1(rp, xbar, std), 2) for rp in rps})
    return ret_pers


def low_return_periods(hist: pd.DataFrame, rps: tuple = (2, 5, 10, 25, 50, 100)) -> dict:
    """
    Solves the Gumbel Type-I distribution using the annual minimum flow from the historic simulation

    Args:
        hist: the csv response from the HistoricSimulation streamflow data service
        rps: a tuple of integer return period numbers to compute

    Returns:
        dictionary with keys labeled f'{return_period}_year' and float values
    """
    annual_min_flow_list = []
    low_flows = {}

    year_min = hist.index.min().year
    year_max = hist.index.max().year

    for y in range(year_min, year_max + 1):
        annual_min_flow_list.append(hist[hist.index.year == int(y)].min())

    annual_min_flow_list = np.array(annual_min_flow_list)
    xbar = np.mean(annual_min_flow_list)
    std = np.std(annual_min_flow_list)

    for rp in rps:
        value = gumbel1(rp, xbar, std)
        value = value if value > 0 else 0
        low_flows[f'return_period_{rp}'] = value

    return low_flows
